/*
File:		$RCSfile$
Version:	$Revision$
Modified By:	$Author$
Last Modified:	$Date$
*/

#include	<vxWorks.h>
#include	<stdioLib.h>
#include	<string.h>
#include        <semLib.h>	/* jps: include for init_record wait */
#include	<epicsDynLink.h>
#include	<sysSymTbl.h>	/* for sysSymTbl*/

#include	<alarm.h>
#include	<callback.h>
#include	<dbDefs.h>
#include	<dbAccess.h>
#include	<dbCommon.h>
#include	<fast_lock.h>
#include	<recSup.h>
#include	<devSup.h>
#include	<drvSup.h>

#include	<motorRecord.h>
#include	"motor.h"
#include	"drvMM4000.h"
#include	"motordevCom.h"

#define STATIC static

#define ABS(f) ((f)>0 ? (f) : -(f))
#define SIGN(f) (long)((f)>0 ? 1 : -1)
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define MIN(a,b) ((a)<(b) ? (a) : (b))

/* ----------------Create the dsets for devMM4000----------------- */
/* static long report(); */
STATIC struct driver_table *drvtabptr;
STATIC long MM4000_init(int);
STATIC long MM4000_init_record(struct motorRecord *);
STATIC long MM4000_start_trans(struct motorRecord *);
STATIC long MM4000_build_trans(motor_cmnd, double *, struct motorRecord *);
STATIC long MM4000_end_trans(struct motorRecord *);

struct motor_dset devMM4000 =
{
    {8, NULL, MM4000_init, MM4000_init_record, NULL},
    motor_update_values,
    MM4000_start_trans,
    MM4000_build_trans,
    MM4000_end_trans
};


/* --------------------------- program data --------------------- */

/* This table is used to define the command types */

static int MM4000_table[] = {
    MOTION, 	/* MOVE_ABS */
    MOTION, 	/* MOVE_REL */
    MOTION, 	/* HOME_FOR */
    MOTION, 	/* HOME_REV */
    IMMEDIATE,	/* LOAD_POS */
    IMMEDIATE,	/* SET_VEL_BASE */
    IMMEDIATE,	/* SET_VELOCITY */
    IMMEDIATE,	/* SET_ACCEL */
    IMMEDIATE,	/* GO */
    IMMEDIATE,	/* SET_ENC_RATIO */
    INFO,	/* GET_INFO */
    MOVE_TERM,	/* STOP_AXIS */
    VELOCITY,	/* JOG */
    IMMEDIATE,	/* SET_PGAIN */
    IMMEDIATE,	/* SET_IGAIN */
    IMMEDIATE,	/* SET_DGAIN */
    IMMEDIATE,	/* SET_CGAIN */
    IMMEDIATE,	/* ENABLE_TORQUE */
    IMMEDIATE,	/* DISABL_TORQUE */
    IMMEDIATE,	/* PRIMITIVE */
    IMMEDIATE,	/* SET_HIGH_LIMIT */
    IMMEDIATE,	/* SET_LOW_LIMIT */
};


static struct board_stat **MM4000_cards;

/* --------------------------- program data --------------------- */


/* initialize device support for MM4000 stepper motor */
STATIC long MM4000_init(int after)
{
    SYM_TYPE type;
    long rtnval;

    if (after == 0)
    {
	rtnval = symFindByNameEPICS(sysSymTbl, "_MM4000_access",
		    (void *) &drvtabptr, &type);
	if (rtnval != OK)
	    return(rtnval);
    /*
	IF before DB initialization.
	    Initialize MM4000 driver (i.e., call init()). See comment in
		drvMM400.c init().
	ENDIF
    */
	(drvtabptr->init)();
    }

    rtnval = motor_init_com(after, *drvtabptr->cardcnt_ptr, drvtabptr, &MM4000_cards);
    return(rtnval);
}


/* initialize a record instance */
STATIC long MM4000_init_record(struct motorRecord *mr)
{
    return(motor_init_record_com(mr, *drvtabptr->cardcnt_ptr, drvtabptr, MM4000_cards));
}


/* start building a transaction */
STATIC long MM4000_start_trans(struct motorRecord *mr)
{
    return(motor_start_trans_com(mr, MM4000_cards, NULL));
}


/* end building a transaction */
STATIC long MM4000_end_trans(struct motorRecord *mr)
{
    return(motor_end_trans_com(mr, drvtabptr));
}


/* add a part to the transaction */
STATIC long MM4000_build_trans(motor_cmnd command, double *parms, struct motorRecord *mr)
{
    struct motor_trans *trans = (struct motor_trans *) mr->dpvt;
    struct mess_node *motor_call;
    struct mess_info *motor_info;
    char buff[30];
    int axis, card;
    int maxdigits = sizeof(buff) - 13;
    double dval;

    buff[0] = '\0';

    motor_call = &(trans->motor_call);
    card = motor_call->card;
    axis = motor_call->signal + 1;

    if (MM4000_table[command] > motor_call->type)
	motor_call->type = MM4000_table[command];

    if (trans->state != BUILD_STATE)
	return(ERROR);

    if (command == PRIMITIVE && mr->init != NULL && strlen(mr->init) != 0)
    {
	strcat(motor_call->message, mr->init);
	strcat(motor_call->message, "\r");
    }

    switch (command)
    {
	case MOVE_ABS:
	case MOVE_REL:
	case HOME_FOR:
	case HOME_REV:
	case JOG:
	    if (strlen(mr->prem) != 0)
	    {
		strcat(motor_call->message, mr->prem);
		strcat(motor_call->message, ";");
	    }
	    if (strlen(mr->post) != 0)
		motor_call->postmsgptr = (char *) &mr->post;
	    break;
        default:
	    break;
    }


    switch (command)
    {
	case MOVE_ABS:
	    dval = parms[0] + trans->offset;
	    sprintf(buff, "%dPA%.*g;", axis, maxdigits, dval);
	    break;
	case MOVE_REL:
	    dval = parms[0];
	    sprintf(buff, "%dPR%.*g;", axis, maxdigits, dval);
	    break;
	case HOME_FOR:
	case HOME_REV:
	    sprintf(buff, "%dOR;", axis);
	    break;
	case LOAD_POS:
	    /* The MM4000 can't be told where it is.  Keep track via an offset */
	    trans->offset = trans->motor_pos - parms[0];
	    break;
	case SET_VEL_BASE:
	    break;	    /* MM4000 does not use base velocity */
	case SET_VELOCITY:
	    dval = parms[0];
	    sprintf(buff, "%dVA%.*g;", axis, maxdigits, dval);
	    break;
	case SET_ACCEL:
	    /* 
	     * The value passed is in steps/sec/sec.  
	     * Convert to user units/sec/sec
	     */
	    dval = parms[0];
	    sprintf(buff, "%dAC%.*g;", axis, maxdigits, dval);
	    break;
	case GO:
	    /* 
	     * The MM4000 starts moving immediately on move commands, GO command
	     * does nothing
	     */
	    break;
	case SET_ENC_RATIO:
	    /*
	     * The MM4000 does not have the concept of encoder ratio, ignore this
	     * command
	     */
	    break;
	case GET_INFO:
	    /* These commands are not actually done by sending a message, but
	       rather they will indirectly cause the driver to read the status
	       of all motors */
	    break;
	case STOP_AXIS:
	    sprintf(buff, "%dST;", axis);
	    break;
	case JOG:
	    /* 
	    * MM4000 does not have a jog command.  Simulate with move absolute
	    * to the appropriate software limit. We can move to MM4000 soft limits.
	    * If the record soft limits are set tighter than the MM4000 limits
	    * the record will prevent JOG motion beyond its soft limits 
	    */
	    dval = parms[0];
	    sprintf(buff, "%dVA%.*g;", axis, maxdigits, ABS(dval));
	    strcat(motor_call->message, buff);
	    if (dval > 0.)
		sprintf(buff, "%dPA%.*g;", axis, maxdigits, mr->dhlm);
	    else
		sprintf(buff, "%dPA%.*g;", axis, maxdigits, mr->dllm);
	    break;
	case SET_CGAIN:
	    sprintf(buff, "%dKP%f;%dKI%f;%dKD%f;%dUF", axis, dval, axis,
		(2.0 * dval), axis, (3.0 * dval), axis);
	    break;
	case SET_PGAIN:
	    sprintf(buff, "%dKP%f;%dUF", axis, (double) parms[0], axis);
	    break;
	case SET_IGAIN:
	    sprintf(buff, "%dKI%f;%dUF;", axis, (double) parms[0], axis);
	    break;
	case SET_DGAIN:
	    sprintf(buff, "%dKD%f;%dUF;", axis, (double) parms[0], axis);
	    break;
	case ENABLE_TORQUE:
	    sprintf(buff, "MO;");
	    break;
	case DISABL_TORQUE:
	    sprintf(buff, "MF;");
	    break;
	case SET_HIGH_LIMIT:
	    motor_info = &(*trans->tabptr->card_array)[card]->motor_info[axis-1];
	    trans->state = IDLE_STATE;	/* No command sent to the controller. */

	    if (*parms > motor_info->high_limit)
	    {
		mr->dhlm = motor_info->high_limit * mr->res;
		return(ERROR);
	    }
	    else
		return(OK);
	case SET_LOW_LIMIT:
	    motor_info = &(*trans->tabptr->card_array)[card]->motor_info[axis-1];
	    trans->state = IDLE_STATE;	/* No command sent to the controller. */

	    if (*parms < motor_info->low_limit)
	    {
		mr->dllm = motor_info->low_limit * mr->res;
		return(ERROR);
	    }
	    else
		return(OK);
	default:
	    return(ERROR);
    }
    strcat(motor_call->message, buff);

    return(OK);
}
